/*	Copyright (C) 2018-2024 Martin Guy <martinwguy@gmail.com>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 3
 *	as published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* libmpg123.c: Interface between spettro and libmpg123 for MP3 decoding
 *
 * Adapted from mpg123's example program feedseek.c, which is copyright 2008
 * by the mpg123 project - free software under the terms of the LGPL 2.1
 */

#include "spettro.h"
#include "libmpg123.h"

#if HAVE_LIBMPG123

#include <string.h>	/* for strerror() */
#include <errno.h>
#include <mpg123.h>
#if HAVE_LIBMAGIC
# include <magic.h>
#endif

/* Open an MP3 file, setting af->{sample_rate,channels,frames}
 * On failure, returns FALSE.
 */
bool
libmpg123_open(audio_file_t *af, char *filename)
{
    int ret;
    static bool initialized = FALSE;

    /* Unfortunately, libmpg123_{open,getformat}() "succeed" in opening
     * some WAV files and some Ogg files but returns the file length as
     * 0.048 seconds and gives garbage data, so do a pre-emptive check
     * on the file type.
     */
#if HAVE_LIBMAGIC
    magic_t magic;
    const char *mime_type;

    magic = magic_open(MAGIC_SYMLINK | MAGIC_MIME_TYPE | MAGIC_ERROR);
    if (magic == NULL) {
	fprintf(stderr, "Error opening libmagic");
	if (magic_error(magic) != NULL)
	    fprintf(stderr, ": %s\n", magic_error(magic));
	else
	    perror("");
	return FALSE;
    }
    if (magic_load(magic, NULL) != 0 ||
        (mime_type = magic_file(magic, filename)) == NULL ||
	strcmp(mime_type, "audio/mpeg") != 0) {
	magic_close(magic);
	return FALSE;
    }

    magic_close(magic);
#else
# warning "Using libmpg123 without libmagic opens some Ogg/WAV files with libmpg123 and returns garbage"
#endif
 
    if (!initialized) {
	mpg123_init();
	initialized = TRUE;
    }

    if ((af->mh = mpg123_new(NULL, &ret)) == NULL) {
	fprintf(stderr,"Unable to create mpg123 handle: %s\n",
		mpg123_plain_strerror(ret));
	return FALSE;
    }

    /* Suppress warning messages from libmpg123 */
    mpg123_param(af->mh, MPG123_ADD_FLAGS, MPG123_QUIET, 0.0);

    if (mpg123_open(af->mh, filename) != MPG123_OK) goto fail_delete;

    /* Set the output format to signed short but keep the sampling rate */
    {
	long rate; int chans; int encoding;

	if (mpg123_getformat(af->mh, &rate, &chans, &encoding) != MPG123_OK)
	    goto fail_close;

	/* Use 16-bit signed output, right for the cache */
	if (mpg123_format_none(af->mh) != MPG123_OK ||
	    mpg123_format(af->mh, rate,  MPG123_MONO | MPG123_STEREO,
					  MPG123_ENC_SIGNED_16) != MPG123_OK)
	    goto fail_close;

	af->channels = chans;
	af->sample_rate = rate;

	/* We may need to call mpg123_scan() to determine the track length
	 * accurately, but that takes ten seconds for an hour-long track
	 * due to disk IO time. However,
	 *
	 > On 23/08/2020, Thomas Orgis <thomas-forum@orgis.org> wrote:
	 > 1. Call mpg123_info() and check for the vbr mode.
	 > If it's not MPG123_CBR, you had some Xing/Lame/Info frame
	 > to tell the decoder that. Track length is a common feature
	 > that all these frames have.
	 >
	 > 2. Call mpg123_getstate() for MPG123_ENC_DELAY and/or
	 > MPG123_ENC_PADDING. If one of them is != -1, you got a
	 > Lame info tag that contained such a value, also implying
	 > that there is proper length info.
	 *
	 * So we do that.
	 */
	{
	    struct mpg123_frameinfo mi;
#if MPG123_API_VERSION >= 45
	    long val;
#endif

	    if (!((mpg123_info(af->mh, &mi) == MPG123_OK &&
		   mi.vbr != MPG123_CBR)
/* These are only present from libmpg123 1.26 */
#if MPG123_API_VERSION >= 45
		   ||
		  (mpg123_getstate(af->mh, MPG123_ENC_DELAY, &val, NULL)
		   == MPG123_OK && val != -1)
		   ||
		  (mpg123_getstate(af->mh, MPG123_ENC_PADDING, &val, NULL)
		   == MPG123_OK && val != -1)
#endif
		   )) (void) mpg123_scan(af->mh);
	}
	{
	    frames_t length = mpg123_length(af->mh);
	    if (length < 0) {
		fprintf(stderr, "Can't discover the length of the MP3 file.\n");
		goto fail_close;
	    }
	    af->frames = length;
	}
    }

    af->filename = filename;

    return TRUE;

fail_close:
    mpg123_close(af->mh);
fail_delete:
    mpg123_delete(af->mh);
    af->mh = NULL;
    return FALSE;
}

/* Seek to the "start"th sample from the MP3 file.
 * Returns TRUE on success, FALSE on failure.
 */
bool
libmpg123_seek(audio_file_t *af, int start)
{
    if (mpg123_seek(af->mh, (frames_t)start, SEEK_SET) == MPG123_ERR) {
	fprintf(stderr, "Failed to seek in MP3 file: %s\n", mpg123_strerror(af->mh));
	return FALSE;
    }

    return TRUE;
}

/*
 * Read samples from the MP3 file and convert them to the desired format
 * into the buffer "write_to".
 *
 * Returns the number of frames written, or a negative value on errors.
 */
int
libmpg123_read_frames(audio_file_t *af,
		      void *write_to,
		      frames_t frames_to_read,
		      af_format_t format)
{
    /* Size of an input sample frame in bytes */
    int framesize = sizeof(short) * af->channels;
    /* Number of sample frames written into write_to */
    int frames_written = 0; /* Quieten "may be used uninitialized" warning */

    /* Avoid reading past end of file because that makes the whole read fail */
    {
	off_t r = mpg123_tell(af->mh);
	if (r < 0) {
	    fprintf(stderr, "libmpg123 can't tell where it is!\n");
	    return -1;
	}

	/* mpg123_tell() returns result in sample frames,
	 * not in samples like it says in the docs */
	if (r + frames_to_read > af->frames) {
	fprintf(stderr, "Reducing read length from %ld ", frames_to_read);
	    frames_to_read = af->frames - r;
	fprintf(stderr, "to %ld\n", frames_to_read);
	}
    }

    switch (format) {
	size_t bytes_to_read, bytes_written;
    case af_signed:
	{
	    bytes_to_read = frames_to_read * framesize;

	    if (mpg123_read(af->mh, write_to, bytes_to_read, &bytes_written)
	        != MPG123_OK) { abort(); return -1; }
	    if (bytes_written != bytes_to_read)
		fprintf(stderr, "mpg123_read() returned %lu of %lu bytes\n",
			(long unsigned) bytes_written,
			(long unsigned) bytes_to_read);

	    frames_written = bytes_written / framesize;
	}
	break;
    case af_float:  
	{
	    signed short *buf = Malloc(frames_to_read * af->channels * sizeof(*buf));
	    bytes_to_read = frames_to_read * framesize;

	    if (mpg123_read(af->mh, (unsigned char *)buf,
			    bytes_to_read, &bytes_written)
	        != MPG123_OK) {
		free(buf);
		return -1;
	    }
	    frames_written = bytes_written / framesize;

	    switch (af->channels) {
		float *fp; signed short *sp; int i;
	    case 1:
		sp = buf; fp = write_to;
		for (i=0; i<frames_written; i++) {
		    *fp++ = *sp++ / (float)32768;
		}
		break;
	    case 2:
		sp = buf; fp = write_to;
		for (i=0; i<frames_written; i++) {
		    *fp++ = (sp[0] + sp[1]) / (float)65536;
		    sp += 2;
		}
		break;
	    }
	    free(buf);
	}
	break;
    }
    return frames_written;
}

void
libmpg123_close(audio_file_t *af)
{
    mpg123_close(af->mh);
    mpg123_delete(af->mh);
}
#endif /* HAVE_LIBMPG123 */
